#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>
//gcc rawgene5.cpp -o rawgene5.exe -lm -Wall -O2 -march=i686 -msse

//       Intel SSE intrinsics: Need Pentium III or Celeron(1999~)
#ifdef __SSE__
#include <xmmintrin.h>
#endif

// #include <emmintrin.h> // SSE2
// #include <pmmintrin.h> // SSE3

#define INITIAL_OFFSET (4410)
#define SEARCH_BACK_WIDTH (4410)
#define SEARCH_FORW_WIDTH (8820)
#define INITIAL_MAX (SEARCH_BACK_WIDTH + SEARCH_FORW_WIDTH)
#define MINIMAL_WAV_LENGTH (INITIAL_OFFSET + SEARCH_BACK_WIDTH + SEARCH_FORW_WIDTH)
//                   17640 = 0.4 sec, wav files of 0.4 sec or shorter is totally invalid.
#define RECOMMENDED_WAV_LENGTH (MINIMAL_WAV_LENGTH+4410)
//                   22050 = 0.5 sec, at least 0.1 sec of valid section is recommended.
#define MAX_ACCEPTABLE_AMP (0.99)


#define WAV_MAX (1+13)
#define PATH_LENGTH_MAX (1000)
#define THUMB_WINDOW (14)

#define OUTPUT_FORMAT_RAW (0)
#define OUTPUT_FORMAT_WAV (1)

#define FILTER_YULEWALK (0)
#define FILTER_FLAT (1)

struct wav_common{
 float *initials;
 double gain;
 double score;//smaller the better
 int sample_rate;
 int sample_num;
 int shortest_sample_num;
 int offset;
 int file_offset;
 int is_test;
 int is_float;
 int is_16bit;
 int use_filter;
 int is_sse_optim;
 int wav_index;
 double max_amp;
 char thumbnail[6+THUMB_WINDOW+1+1]; 
 FILE WAV;
 FILE* wav;
};

int init_wav_common(struct wav_common *wc){
 wc->gain = 1.0;
 wc->score = 0.0;
 wc->sample_rate = 0;
 wc->sample_num = 0;
 wc->shortest_sample_num = 0;
 wc->offset = 0;
 wc->file_offset = 0;
 wc->wav_index = 9;
 wc->is_test = 0;
 wc->is_float = 0;
 wc->is_16bit = 0;
 wc->use_filter = FILTER_YULEWALK;
 wc->is_sse_optim = 1;
 wc->max_amp = 0.0;
 strcpy(wc->thumbnail,"??? ?>**************>");
 wc->wav = &(wc->WAV);
 return 0;
}

 /*------------------------------------------------------------------*/
 /*             Brute-force search of offset and gain                */
 /*              compare Wav-Original(wo) and Wav-Compare(wc)        */
 /*               results are stored in Wav-Compare(wc)              */
 /*                The most time-consuming process, typically 90%.   */
 /*------------------------------------------------------------------*/

int wav_gainoffset_search(struct wav_common *wo,struct wav_common *wc){
 int i,j;
 int offset;
 int offset_end;
 double bestscore = 999999999999999.9;
 double score;
 double bestamp = 0.0;
 int bestoffset = 0;
 //typical offset: 0 1088 1 576 287 
 double rmsampo = 0.0;  //  Root Mean Square AMPlitude of Original
 double rmsampc = 0.0;  //  Root Mean Square AMPlitude of Compare
 double error_wav;
 int theend = wo->shortest_sample_num - INITIAL_OFFSET - SEARCH_FORW_WIDTH;
 for(i=SEARCH_BACK_WIDTH;i<theend;i++){
  double temp = wo->initials[i];
  rmsampo += temp*temp;
 }
 if(wo->sample_rate!=44100 && wo->sample_rate!=48000)printf("WO->sample_rate = %d Hz error.\n",wo->sample_rate);
 if(wo->sample_rate!=wc->sample_rate)printf("WO->sample_rate(%d)!=WC->sample_rate(%d) error.\n",wo->sample_rate,wc->sample_rate);
 if(rmsampo!=rmsampo || rmsampo<=0.0){
   printf("Error: the original wav file is silent or collapsed.\n theend=%d rmsampo=%f \r\n",theend,rmsampo);
 }
 for(j=0;j<=3;j++){
  //offset(encoder delay) is likely to be
  // 0, 1, 287, 576, 1024, 1025, 1057, 1088, 1105, 1600, 2048, 2209, 6049.
  // in typical MP3/AAC encoders.
  if(j==0){//-8 ~ 2342
   offset = -(SEARCH_BACK_WIDTH>>9);
   offset_end = (SEARCH_FORW_WIDTH>>2) + (SEARCH_FORW_WIDTH>>6);
  }else if(j==1){//-551 ~ -8
   offset = -(SEARCH_BACK_WIDTH>>3);
   offset_end = -(SEARCH_BACK_WIDTH>>9);
  }else if(j==2){//2342 ~ 8820
   offset = (SEARCH_FORW_WIDTH>>2) + (SEARCH_FORW_WIDTH>>6);
   offset_end = SEARCH_FORW_WIDTH;
  }else if(j==3){//-4410 ~ -551
   offset = -SEARCH_BACK_WIDTH;
   offset_end = -(SEARCH_BACK_WIDTH>>3);
  }
  //printf("offset=%d - %d\n",offset,offset_end);
  if(offset>offset_end)printf("OFFSET_ERROR,offset %d> offset_end %d \r\n",offset,offset_end);
  // search these areas first, to minimize execution time.
  rmsampc = 0.0;
  for(i=SEARCH_BACK_WIDTH;i<theend;i++){
   double temp = wc->initials[offset+i];
   rmsampc += temp*temp;
  }
  for(;offset<offset_end;offset++){
   if(rmsampc!=rmsampc || rmsampc<=0.0){
    printf("Error: No Sound At All. rmsampc=%f \r\n",rmsampc);
    return 0;
   }
   double theamp = sqrt(rmsampo/rmsampc);
   score = 0.0;
   i = SEARCH_BACK_WIDTH;
#ifdef __SSE__
   if(wc->is_sse_optim){//using SSE code.
     //Intel SSE instructions
     //__attribute__((aligned(16))) float score_f[4];
     float score_f;
     __m128 theamp_p = _mm_set1_ps((float)theamp);
     while(i+16<=theend){
       __m128 temp;
       __m128 initials_p;
       __m128 sum_p;
       temp = _mm_loadu_ps(&(wc->initials[offset+i]));
       temp = _mm_mul_ps(temp,theamp_p);
       initials_p = _mm_loadu_ps(&(wo->initials[i]));
       temp = _mm_sub_ps(temp,initials_p);
       temp = _mm_mul_ps(temp,temp);
       sum_p = temp;
       i += 4;
       temp = _mm_loadu_ps(&(wc->initials[offset+i]));
       temp = _mm_mul_ps(temp,theamp_p);
       initials_p = _mm_loadu_ps(&(wo->initials[i]));
       temp = _mm_sub_ps(temp,initials_p);
       temp = _mm_mul_ps(temp,temp);
       sum_p = _mm_add_ps(sum_p,temp);
       i += 4;
       temp = _mm_loadu_ps(&(wc->initials[offset+i]));
       temp = _mm_mul_ps(temp,theamp_p);
       initials_p = _mm_loadu_ps(&(wo->initials[i]));
       temp = _mm_sub_ps(temp,initials_p);
       temp = _mm_mul_ps(temp,temp);
       sum_p = _mm_add_ps(sum_p,temp);
       i += 4;
       temp = _mm_loadu_ps(&(wc->initials[offset+i]));
       temp = _mm_mul_ps(temp,theamp_p);
       initials_p = _mm_loadu_ps(&(wo->initials[i]));
       temp = _mm_sub_ps(temp,initials_p);
       temp = _mm_mul_ps(temp,temp);
       sum_p = _mm_add_ps(sum_p,temp);
       i += 4;
       //_mm_store_ps(score_f,sum_p);
       //score += (score_f[0]+score_f[1]+score_f[2]+score_f[3]);
       sum_p = _mm_add_ps(sum_p, _mm_movehl_ps(temp,sum_p));
       sum_p = _mm_add_ss(sum_p, _mm_shuffle_ps(sum_p, sum_p, 1));
       _mm_store_ss(&score_f,sum_p);
       score += score_f;
       if(score>bestscore)break;
       //if(offset+i<0 || offset+i>=(wo->shortest_sample_num - INITIAL_OFFSET))printf("\r\nERROR\r\n");
       //if(i<0 || i>=(wo->shortest_sample_num - INITIAL_OFFSET))printf("\r\nERROR\r\n");
     }
   }else{
#endif
     //NO SSE OPTIMIZATION for OLD CPUs.
     while(i+16<=theend){
       error_wav = (wc->initials[offset+i]*theamp)-wo->initials[i];
       score += error_wav*error_wav;
       i++;
       error_wav = (wc->initials[offset+i]*theamp)-wo->initials[i];
       score += error_wav*error_wav;
       i++;
       error_wav = (wc->initials[offset+i]*theamp)-wo->initials[i];
       score += error_wav*error_wav;
       i++;
       error_wav = (wc->initials[offset+i]*theamp)-wo->initials[i];
       score += error_wav*error_wav;
       i++;
       error_wav = (wc->initials[offset+i]*theamp)-wo->initials[i];
       score += error_wav*error_wav;
       i++;
       error_wav = (wc->initials[offset+i]*theamp)-wo->initials[i];
       score += error_wav*error_wav;
       i++;
       error_wav = (wc->initials[offset+i]*theamp)-wo->initials[i];
       score += error_wav*error_wav;
       i++;
       error_wav = (wc->initials[offset+i]*theamp)-wo->initials[i];
       score += error_wav*error_wav;
       i++;
       error_wav = (wc->initials[offset+i]*theamp)-wo->initials[i];
       score += error_wav*error_wav;
       i++;
       error_wav = (wc->initials[offset+i]*theamp)-wo->initials[i];
       score += error_wav*error_wav;
       i++;
       error_wav = (wc->initials[offset+i]*theamp)-wo->initials[i];
       score += error_wav*error_wav;
       i++;
       error_wav = (wc->initials[offset+i]*theamp)-wo->initials[i];
       score += error_wav*error_wav;
       i++;
       error_wav = (wc->initials[offset+i]*theamp)-wo->initials[i];
       score += error_wav*error_wav;
       i++;
       error_wav = (wc->initials[offset+i]*theamp)-wo->initials[i];
       score += error_wav*error_wav;
       i++;
       error_wav = (wc->initials[offset+i]*theamp)-wo->initials[i];
       score += error_wav*error_wav;
       i++;
       error_wav = (wc->initials[offset+i]*theamp)-wo->initials[i];
       score += error_wav*error_wav;
       i++;
       if(score>bestscore)break;
     }
#ifdef __SSE__
   }
#endif

   while(i<theend){
    if(score>bestscore)break;
    error_wav = (wc->initials[offset+i]*theamp)-wo->initials[i];
    score += error_wav*error_wav;
    if(offset+i<0 || offset+i>=(wo->shortest_sample_num - INITIAL_OFFSET))printf("\r\nERROR\r\n");
    if(i<0 || i>=(wo->shortest_sample_num - INITIAL_OFFSET))printf("\r\nERROR\r\n");
    i++;
   }
   if(score<bestscore){
    bestscore = score;
    bestamp = theamp;
    bestoffset = offset;
   }
   rmsampc -= (wc->initials[offset+SEARCH_BACK_WIDTH])*(wc->initials[offset+SEARCH_BACK_WIDTH]);
   rmsampc += (wc->initials[offset+theend])*(wc->initials[offset+theend]);
  }
 }
 wc->gain = bestamp;
 wc->offset = bestoffset;
 wc->score = sqrt(bestscore/rmsampo);
 return 0;
}

 /*------------------------------------------------------------------*/
 /*   Wav File Validity Check(Must be 32/16bit uncompressed stereo)  */
 /*------------------------------------------------------------------*/

int wav_validity_check(struct wav_common *wc, char *wav_path){
  char chunkstr[8] = "";
  char t_nail[6+THUMB_WINDOW+1+1] = "";
  int chunk_length;
  int chunk_max = 10;
  int j;
 
  if(wav_path[0]=='\0'){
   printf("Cannot open this empty path ''\r\n");
   return 2;
  }
  wc->wav = fopen(wav_path,"rb");
  if(wc->wav==NULL){
   printf("Cannot open this file : %s\r\n",wav_path);
   return 2;
  }

  sprintf(t_nail,"??? %01d>",wc->wav_index);
  char *p;//full path to filename routine
  p = strrchr(wav_path,'\\');//search the last backslash
  if(p==NULL){
   p = wav_path;
  }
  else{
   p++;
  }
  strncat(t_nail,p,THUMB_WINDOW);
  while(strlen(t_nail)<6+THUMB_WINDOW)strcat(t_nail," ");
  strcat(t_nail,">");
  fread(chunkstr,1,4,wc->wav);
  if(strcmp(chunkstr,"RIFF")!=0){
   printf("RIFF != %s\r\n",chunkstr);
   printf("%s Not a valid RIFF WAV File: %s\r\n",t_nail,wav_path);
   return 3;
  }
  fread(chunkstr,1,4,wc->wav);
  fread(chunkstr,1,4,wc->wav);
  if(strcmp(chunkstr,"WAVE")!=0)printf("WAVE != %s\r\n",chunkstr);
  fread(chunkstr,1,4,wc->wav);
  while(strcmp(chunkstr,"fmt ")!=0 && chunk_max){
   printf("%s skipped unknown chunk '%s' while searching 'fmt '.\r\n",t_nail,chunkstr);
   fread(&chunk_length,4,1,wc->wav);
   for(j=0;j<chunk_length;j++)fread(chunkstr,1,1,wc->wav);
   fread(chunkstr,1,4,wc->wav);
   chunk_max--;
  }
  if(chunk_max==0){
   printf("%s Chunks too much.\r\n",t_nail);
   return 3;
  }
  fread(&chunk_length,4,1,wc->wav);
  fread(chunkstr,1,4,wc->wav);
  int wav_type = (chunkstr[1]<<8)|chunkstr[0];
  if(wav_type==3){
   wc->is_float = 1;  //Microsoft float 32bit wav file
   t_nail[0] = 'F';
   t_nail[1] = 'L';
   t_nail[2] = 'O';
  }
  else if(wav_type==1){
   wc->is_float = 0;  //Normal wav file
   t_nail[0] = 'I';
   t_nail[1] = 'N';
   t_nail[2] = 'T';
  }
  else if(wav_type==-2){
   wc->is_float = 2;  //WAVE_FORMAT_EXTENSIBLE
   t_nail[0] = 'E';
   t_nail[1] = 'X';
   t_nail[2] = 'T';
  }else{
   printf("%s This is not a linear(integer) or float PCM. %d\r\n",t_nail,wav_type);
   return 4;//not linear or float PCM
  }

  if(chunkstr[3]!=0 || chunkstr[2]!=2){
   printf("%s This is not a stereo file. # of channel is %d.\r\n",t_nail,chunkstr[2]);
   return 5;//channel number error. mono, surround etc.
  }
  int input_sampling_rate;
  int intbuf;

  fread(&input_sampling_rate,1,4,wc->wav);
  wc->sample_rate = input_sampling_rate;
  if(input_sampling_rate!=44100 && input_sampling_rate!=48000){
   printf("%s The sample rate of this file is %dHz.\r\n",t_nail,input_sampling_rate);
   printf("%s Using sample rate other than 44.1kHz, 48kHz is not recommended.\r\n",t_nail);
  }
  fread(&intbuf,1,4,wc->wav);
  if(intbuf!=input_sampling_rate*8){
   if(intbuf!=input_sampling_rate*4){
    printf("%s This is not a 32bit nor 16bit depth wave file.\r\n",t_nail);
    printf("%s This is probably a %d bit wave file.\r\n",t_nail,4*intbuf/input_sampling_rate);
    return 6;//8bit,24bit...etc
   }
   wc->is_16bit = 1;
   t_nail[1] = 'n';
   t_nail[2] = 't';
  }
 
  if(wc->is_float==2){
   for(j=0;j<12;j++)fread(chunkstr,1,1,wc->wav);
   fread(chunkstr,1,4,wc->wav);
   if(((chunkstr[1]<<8)|chunkstr[0]) == 3){
    //wc->is_float = 1; //Microsoft float 32bit wav file
    t_nail[0] = 'F';
    t_nail[1] = 'L';
    t_nail[2] = 'X';
   }else{
    wc->is_float = 0;
    t_nail[0] = 'I';
    t_nail[1] = 'N';
    t_nail[2] = 'X';
   }
   for(j=0;j<chunk_length-28;j++)fread(chunkstr,1,1,wc->wav);
  }else{
   for(j=0;j<chunk_length-12;j++)fread(chunkstr,1,1,wc->wav);
  }
  strcpy(wc->thumbnail,t_nail);
  //printf("%d fmt complete\r\n",i);
  fread(chunkstr,1,4,wc->wav);
  while(strcmp(chunkstr,"data")!=0 && chunk_max){
   //printf("%s skipped unknown chunk '%s' while searching 'data'.\r\n",t_nail,chunkstr);
   fread(&chunk_length,4,1,wc->wav);
   if(chunk_length>=0x10000)printf("%s Very large chunk name:%s size:%x \r\n",t_nail,chunkstr,chunk_length);
   for(j=0;j<chunk_length;j++)fread(chunkstr,1,1,wc->wav);
   fread(chunkstr,1,4,wc->wav);
   chunk_max--;
  }
  if(chunk_max==0){
   printf("%s There are no data chunk.\r\n",t_nail);
   return 7;
  }
  fread(&intbuf,4,1,wc->wav);
  if(wc->is_16bit){
   if(intbuf&0x03)printf("%s ILLEGAL DATA CHUNK LENGTH : %d\r\n",t_nail,intbuf);
   intbuf /= 4;
  }
  else{
   if(intbuf&0x07)printf("%s ILLEGAL DATA CHUNK LENGTH : %d\r\n",t_nail,intbuf);
   intbuf /= 8;
  }
  wc->sample_num = intbuf;
  wc->file_offset = (int)ftell(wc->wav);
  return 0;
}

int pick_yulewalked_samples(struct wav_common *wc){

  /*------------------------------------------------------------------*/
  /*     Cut inaudible sounds by Yulewalk filter,                     */
  /*wiki.hydrogenaudio.org/index.php?title=ReplayGain_1.0_specification*/
  /*      and store INITIAL_OFFSET~shortest sample num section        */
  /*       of the sound for further analysis.                         */
  /*     Also check peak amplitude of the sound to avoid clipping.    */
  /*------------------------------------------------------------------*/
  //"Yulewalk" filter
  double iir_l_left[10];
  double iir_l_right[10];
  double iir_b_left[10];
  double iir_b_right[10];
  double dbuf;
  float flobuf;
  int intbuf;
  short shobuf;
  double cap1,cap0;
  int k,j;
  int sample_num = wc->sample_num;
  int shortest_sample_num = wc->shortest_sample_num;
  int wav_index = wc->wav_index;
  int is_test = wc->is_test;
  double hz_angle (2*3.1415926/(wc->sample_rate));
  double the_amp;
  double max_amp = 0.0;
  static int debug_remain = 11;
  for(k=0;k<10;k++){
   iir_l_left[k] = iir_l_right[k] = 0;
   iir_b_left[k] = iir_b_right[k] = 0;
  }
  wc->initials = (float *)calloc(shortest_sample_num-INITIAL_OFFSET+1 , sizeof(float));
  if(wc->initials == NULL){
   printf("Wav file too long, memory too few.\r\n");
   return 10;
  }
  //printf("wc->sample_rate = %d \n",wc->sample_rate);
  //printf("wc->is_float=%d wc->is_test=%d wc->is_16bit=%d \n",wc->is_float, wc->is_test, wc->is_16bit);
 
  if(!is_test)fseek(wc->wav,wc->file_offset,SEEK_SET);

  for(j=0;j<sample_num;j++){
  /*------------------------------------------------------------------*/
  /*     Read (or generate) sound sample.                             */
  /*      Right and left channel will be combined as mono.            */
  /*------------------------------------------------------------------*/   
   if(wc->is_float){
    if(wc->is_test){//If it's self test mode, generate test waveform.
     const int delay_num[10] = {0,-1,23,-456,7890,-987,65,-4,3,-2};
     k = j-delay_num[(wav_index<10 && 0<=wav_index)?wav_index:0];
     //Audible 678Hz 1.234kHz 5.678kHz

     the_amp = cos(hz_angle*678*k)+sin(hz_angle*1234.0*k)+cos(hz_angle*5678.9*k);
     //Add Inaudible sounds to test if they confuse the gain normalizer.
     if(wav_index%3==0)the_amp += sin(hz_angle*20000*k);
     if(wav_index%2==0)the_amp += cos(hz_angle*12.3*k);
     the_amp /= (10.0+wav_index);
     if(fabs(the_amp)>max_amp)max_amp = fabs(the_amp);
     dbuf = the_amp;     
    }
    else{// read from 32bit float wav file.
     fread(&flobuf,4,1,wc->wav);
     the_amp = (double)flobuf;
     dbuf = the_amp;
     if(fabs(the_amp)>max_amp)max_amp = fabs(the_amp);
     if(debug_remain>=0 && flobuf!=flobuf){
      printf("The float wav file is corrupted, %f found. (%d) \r\n",flobuf,debug_remain);debug_remain--;
      printf("wc->is_float=%d wc->is_test=%d wc->is_16bit=%d \n",wc->is_float, wc->is_test, wc->is_16bit);
     }
     fread(&flobuf,4,1,wc->wav);
     the_amp = (double)flobuf;
     dbuf += the_amp;
     if(fabs(the_amp)>max_amp)max_amp = fabs(the_amp);
     if(debug_remain>=0 && flobuf!=flobuf){
      printf("The float wav file is corrupted, %f found. (%d) \r\n",flobuf,debug_remain);debug_remain--;
      printf("wc->is_float=%d wc->is_test=%d wc->is_16bit=%d \n",wc->is_float, wc->is_test, wc->is_16bit);
     }
    }
   }
   else{
    if(wc->is_16bit){//read from 16bit normal integer wav file.
     fread(&shobuf,2,1,wc->wav);
     the_amp = ((double)shobuf)/32768.0;
     dbuf = the_amp;
     if(fabs(the_amp)>max_amp)max_amp = fabs(the_amp);
     fread(&shobuf,2,1,wc->wav);
     the_amp = ((double)shobuf)/32768.0;
     dbuf += the_amp;
     if(fabs(the_amp)>max_amp)max_amp = fabs(the_amp);
    }
    else{// read from 32bit integer wav file.
     fread(&intbuf,4,1,wc->wav);
     the_amp = ((double)intbuf)/2147483648.0;//32bit 2^31
     dbuf = the_amp;
     if(fabs(the_amp)>max_amp)max_amp = fabs(the_amp);
     fread(&intbuf,4,1,wc->wav);
     the_amp = ((double)intbuf)/2147483648.0;
     dbuf += the_amp;
     if(fabs(the_amp)>max_amp)max_amp = fabs(the_amp);
    }
   }
  /*------------------------------------------------------------------*/
  /*     Filter. Equalize. Inaudible sounds will be totally cut off.  */
  /*      Skip first 0.1sec of the filtered sound for accuracy.       */
  /*------------------------------------------------------------------*/
   if(wc->sample_rate <= 46000 && wc->use_filter==FILTER_YULEWALK){//44.1kHz
     cap0 = dbuf*0.98500175787242;
     cap0 -= iir_b_left[0]*1.97000351574484;
     cap0 += iir_b_left[1]*0.98500175787242;
     iir_b_left[1] = iir_b_left[0];
     iir_b_left[0] = dbuf;
     cap0 += iir_b_right[0]*1.96977855582618;
     cap0 -= iir_b_right[1]*0.97022847566350;
     iir_b_right[1] = iir_b_right[0];
     iir_b_right[0] = cap0;

     cap1 = cap0*0.05418656406430;

     cap1 -= iir_l_left[0]*0.02911007808948;
     cap1 -= iir_l_left[1]*0.00848709379851;
     cap1 -= iir_l_left[2]*0.00851165645469;
     cap1 -= iir_l_left[3]*0.00834990904936;
     cap1 += iir_l_left[4]*0.02245293253339;
     cap1 -= iir_l_left[5]*0.02596338512915;
     cap1 += iir_l_left[6]*0.01624864962975;
     cap1 -= iir_l_left[7]*0.00240879051584;
     cap1 += iir_l_left[8]*0.00674613682247;
     cap1 -= iir_l_left[9]*0.00187763777362;
     iir_l_left[9] = iir_l_left[8];
     iir_l_left[8] = iir_l_left[7];
     iir_l_left[7] = iir_l_left[6];
     iir_l_left[6] = iir_l_left[5];
     iir_l_left[5] = iir_l_left[4];
     iir_l_left[4] = iir_l_left[3];
     iir_l_left[3] = iir_l_left[2];
     iir_l_left[2] = iir_l_left[1];
     iir_l_left[1] = iir_l_left[0];
     iir_l_left[0] = cap0;
     cap1 += iir_l_right[0]*3.47845948550071;
     cap1 -= iir_l_right[1]*6.36317777566148;
     cap1 += iir_l_right[2]*8.54751527471874;
     cap1 -= iir_l_right[3]*9.47693607801280;
     cap1 += iir_l_right[4]*8.81498681370155;
     cap1 -= iir_l_right[5]*6.85401540936998;
     cap1 += iir_l_right[6]*4.39470996079559;
     cap1 -= iir_l_right[7]*2.19611684890774;
     cap1 += iir_l_right[8]*0.75104302451432;
     cap1 -= iir_l_right[9]*0.13149317958808;

   }
   else if(wc->use_filter==FILTER_YULEWALK){//Yulewalk filter for 48kHz
     cap0 = dbuf*0.98621192462708;
     cap0 -= iir_b_left[0]*1.97242384925416;
     cap0 += iir_b_left[1]*0.98621192462708;
     iir_b_left[1] = iir_b_left[0];
     iir_b_left[0] = dbuf;
     cap0 += iir_b_right[0]*1.97223372919527;
     cap0 -= iir_b_right[1]*0.97261396931306;
     iir_b_right[1] = iir_b_right[0];
     iir_b_right[0] = cap0;

     cap1 = cap0*0.03857599435200;

     cap1 -= iir_l_left[0]*0.02160367184185 ;
     cap1 -= iir_l_left[1]*0.00123395316851;
     cap1 -= iir_l_left[2]*0.00009291677959;
     cap1 -= iir_l_left[3]*0.01655260341619;
     cap1 += iir_l_left[4]*0.02161526843274;
     cap1 -= iir_l_left[5]*0.02074045215285 ;
     cap1 += iir_l_left[6]*0.00594298065125 ;
     cap1 += iir_l_left[7]*0.00306428023191 ;
     cap1 += iir_l_left[8]*0.00012025322027;
     cap1 += iir_l_left[9]*0.00288463683916;
     iir_l_left[9] = iir_l_left[8];
     iir_l_left[8] = iir_l_left[7];
     iir_l_left[7] = iir_l_left[6];
     iir_l_left[6] = iir_l_left[5];
     iir_l_left[5] = iir_l_left[4];
     iir_l_left[4] = iir_l_left[3];
     iir_l_left[3] = iir_l_left[2];
     iir_l_left[2] = iir_l_left[1];
     iir_l_left[1] = iir_l_left[0];
     iir_l_left[0] = cap0;
     cap1 += iir_l_right[0]*3.84664617118067;
     cap1 -= iir_l_right[1]*7.81501653005538;
     cap1 += iir_l_right[2]*11.34170355132042;
     cap1 -= iir_l_right[3]*13.05504219327545;
     cap1 += iir_l_right[4]*12.28759895145294;
     cap1 -= iir_l_right[5]*9.48293806319790;
     cap1 += iir_l_right[6]*5.87257861775999;
     cap1 -= iir_l_right[7]*2.75465861874613;
     cap1 += iir_l_right[8]*0.86984376593551;
     cap1 -= iir_l_right[9]*0.13919314567432;
   }else if(wc->use_filter==FILTER_FLAT){
     cap1 = dbuf;
   }else{
     cap1 = dbuf;
     printf("Error: Unknown filter:%d \r\n",wc->use_filter);
   }

   iir_l_right[9] = iir_l_right[8];
   iir_l_right[8] = iir_l_right[7];
   iir_l_right[7] = iir_l_right[6];
   iir_l_right[6] = iir_l_right[5];
   iir_l_right[5] = iir_l_right[4];
   iir_l_right[4] = iir_l_right[3];
   iir_l_right[3] = iir_l_right[2];
   iir_l_right[2] = iir_l_right[1];
   iir_l_right[1] = iir_l_right[0];
   iir_l_right[0] = cap1;
   /*------------------------------------------------------------------*/
   /*   Store the yulewalked mono sound in huge malloced float array.  */
   /*------------------------------------------------------------------*/  
   if(INITIAL_OFFSET<=j && j<shortest_sample_num){
    int ptr = j-INITIAL_OFFSET;
    wc->initials[ptr] = (float)cap1;
    if(cap1!=cap1 && debug_remain>=0){
      debug_remain--;
      printf("cap1=%f j=%d ptr=%d shobuf=%d debug_remain=%d \r\n",cap1,j,ptr,shobuf,debug_remain);
      printf("wc->is_float=%d wc->is_test=%d wc->is_16bit=%d \n",wc->is_float, wc->is_test, wc->is_16bit);
      for(k=0;k<10;k++){
          printf("iir_l_left[%d]=%f, ",k,iir_l_left[k]);
          printf("iir_l_right[%d]=%f, %d \r\n",k,iir_l_right[k],j);
      }
    }
   }
  }
  wc->max_amp = max_amp;
  if(!is_test && ferror(wc->wav)!=0){
   printf("%s file read error.\r\n",wc->thumbnail);
   return 11;
  }
  return 0;
}

 /*------------------------------------------------------------------*/
 /*             Main routine                                         */
 /*------------------------------------------------------------------*/
int main(int argc, char **argv){
 char wav_path[WAV_MAX][PATH_LENGTH_MAX];
 int i,j;
 int shortest_sample_num = 0x7FFFFFFF;
 int testmode = 0;
 int wav_num;
 int error_code;
 int output_format = OUTPUT_FORMAT_RAW;
 int use_filter = FILTER_YULEWALK;
 int is_sse_optim = 1;

 struct wav_common WAVC[WAV_MAX];
 struct wav_common *wavc[WAV_MAX];

 wav_num = argc-1;
 if(wav_num>WAV_MAX){
  printf("Too many lossy files : wav_num:%d.\r\n",wav_num);
  return 1;
 }

 if(argc<3){
  printf("Abc/hr raw generator ver.5.0.0 (http://d.hatena.ne.jp/kamedo2/)\r\n");
  printf(" Equalize offset and gain of wav files.\r\n");
  printf("  Accept 32bit float, 32bit, 16bit int 44100/48000 Hz stereo wav.\r\n");
  printf("   Then output them as 16bit raw/wav, to perform an abc/hr test.\r\n\r\n");
  printf("     usage:  rawgene5 original.wav lossy1.wav lossy2.wav.....\r\n\r\n");
  printf(" Detectable offset: %d ~ %d[sample].\r\n",-SEARCH_BACK_WIDTH,SEARCH_FORW_WIDTH-1);
  printf("   Yulewalk filter for 44.1/48kHz is used to normalize gain.\r\n\r\n");
  printf(" Options\r\n");
  printf("  -flat : disable Yulewalk filter and use Unweighted filter in ABC/HR for Java.\r\n");
  printf("  -wav  : output to wav format, instead of raw.\r\n");
  printf("  -test : self-testing the yulewalk filter based gain equalizing.\r\n");
  printf("  -nosse: disable the SSE optimization. Recommended only for very old CPUs.\r\n");
  printf("          the result should be identical.\r\n\r\n");

  if(argc==1)return 0;
  testmode = 1;
 }


 // set initial value, and check validity of all input wav files.

 j = 0;
 for(i=0;i<wav_num;i++){
   //if(!testmode)printf("%d, %d parsing %s... \r\n",i,j,argv[1+i]);
  if(1+i<argc && argv[1+i][0]=='-'){
    if(argv[1+i][1]=='w' && argv[1+i][2]=='a' && argv[1+i][3]=='v'){
      output_format = OUTPUT_FORMAT_WAV;
      printf("argv[1+%d]=-wav, setting to wav output. \r\n",i);
    }else if(argv[1+i][1]=='f' && argv[1+i][2]=='l' && argv[1+i][3]=='a' && argv[1+i][4]=='t'){
      use_filter = FILTER_FLAT;
      printf("argv[1+%d]=-flat, using the flat filter, instead of the yulewalk filter. \r\n",i);
    }else if(argv[1+i][1]=='t' && argv[1+i][2]=='e' && argv[1+i][3]=='s' && argv[1+i][4]=='t'){
      testmode = 1;
      printf("argv[1+%d]=-test, setting the test mode. Test the sanity of the yulewalk filter. \r\n",i);
    }else if(argv[1+i][1]=='n' && argv[1+i][2]=='o' && argv[1+i][3]=='s' && argv[1+i][4]=='s'){
      is_sse_optim = 0;
      printf("argv[1+%d]=-nosse, disabling the sse optimized code. \r\n",i);
    }else{
      printf("argv[1+%d]=%s, Unrecognized option. \r\n",i,argv[1+i]);
    }
  }
 }
 if(testmode){
   wav_num = 8;
   printf("Test mode: self-test the sanity of the volume equalization,\r\n");
   printf(" using internally generated test waveform.\r\n");
   printf("If the default yulewalk filter is properly working, \r\n");
   printf(" the offset shoule be -1, 23, -456, 7890, -987, 65, -4, \r\n");
   printf(" the gain should be about 1.10, 1.20, 1.30, 1.40, 1.50, 1.60, 1.70.\r\n"); 

 }

 for(i=0;i<wav_num;i++){
   if(testmode || (1+i<argc && argv[1+i][0]!='-')){
    wavc[j] = &WAVC[j];
    init_wav_common(wavc[j]);
    wavc[j]->wav_index = j;
    wavc[j]->use_filter = use_filter;
    wavc[j]->is_sse_optim = is_sse_optim;
    if(testmode){
      wavc[j]->is_test = 1;
      wavc[j]->is_float = 1;
      wavc[j]->sample_num = 44100;
      wavc[j]->sample_rate = 44100;
    }
    else{
      if(j==0)printf("rawgene v.5.0.0  /  original reference file:%s\r\n",argv[1+i]);
      strncpy(wav_path[j],argv[1+i],PATH_LENGTH_MAX);
      wav_path[j][PATH_LENGTH_MAX-20] = '\0';
      error_code = wav_validity_check(wavc[j],wav_path[j]);
      if(error_code)return error_code;
    }
    int sample_num = wavc[j]->sample_num;
    if(sample_num<shortest_sample_num){
      shortest_sample_num = sample_num;
      int shortest_wav_length = INITIAL_OFFSET + shortest_sample_num;
      if(shortest_wav_length<RECOMMENDED_WAV_LENGTH){
        printf("MINIMAL_WAV_LENGTH : %d[samples]\r\n",MINIMAL_WAV_LENGTH);
        printf("       RECOMMENDED : %d[samples]\r\n",RECOMMENDED_WAV_LENGTH);
        printf("shortest_wav_length: %d[samples]\r\n",shortest_wav_length);
        printf("%s INPUT FILES TOO SHORT.\r\n",wavc[j]->thumbnail); 
        return 8;
      }
    }
    if(wavc[j]->sample_rate != wavc[0]->sample_rate){
      printf("%s This file has a sampling rate %d, original %d.\r\n",wavc[j]->thumbnail,wavc[j]->sample_rate,wavc[0]->sample_rate);
      return 9;
    }
    j++;
  }//else
 }//for
 wav_num = j;
 j = 0;
 if(wav_num==0){
   printf("Error: There were no arguments of wav files. \r\n");
   printf(" You need 1 original wav file and 1+ comare wav file(s). \r\n");
   return 9;
 }else if(wav_num==1){
   printf("There were a reference wav file %s, but no compare wav file.(lossy1.wav) \r\n", wav_path[0]);
 }

 /*------------------------------------------------------------------*/
 /*  Pick the waves, yulewalk them, and store them in large mallocs. */
 /*------------------------------------------------------------------*/

 wavc[0]->shortest_sample_num = shortest_sample_num;
 error_code = pick_yulewalked_samples(wavc[0]);
 if(error_code)return error_code;

 printf("                    Offset[sample]  Gain   Gain[dB]  Score  MaxAmplitude\r\n");

 int biggest_delay = 0;
 int smallest_delay = 0;
 int usable_length = shortest_sample_num;
 double global_gain = 1.0;
 if(wavc[0]->max_amp>MAX_ACCEPTABLE_AMP)global_gain = MAX_ACCEPTABLE_AMP/wavc[0]->max_amp;
 /*------------------------------------------------------------------*/
 /*     Execute wav_gainoffset_search(Brute-force offset search)     */
 /*------------------------------------------------------------------*/
 for(i=1;i<wav_num;i++){
  wavc[i]->shortest_sample_num = shortest_sample_num;
  error_code = pick_yulewalked_samples(wavc[i]);
  if(error_code)return error_code;

  wav_gainoffset_search(wavc[0],wavc[i]);
  free(wavc[i]->initials);
  wavc[i]->initials = NULL;
  if(wavc[i]->offset<smallest_delay)smallest_delay = wavc[i]->offset;
  if(biggest_delay<wavc[i]->offset)biggest_delay = wavc[i]->offset;
  if(usable_length>wavc[i]->sample_num-wavc[i]->offset)usable_length = wavc[i]->sample_num-wavc[i]->offset;
  if(global_gain*wavc[i]->gain*wavc[i]->max_amp > MAX_ACCEPTABLE_AMP){
   global_gain = MAX_ACCEPTABLE_AMP/(wavc[i]->gain*wavc[i]->max_amp);
  }
  printf("%s",wavc[i]->thumbnail);
  printf("  %8d",wavc[i]->offset);
  printf("  %9f",wavc[i]->gain);
  printf(" %9f",20.0*log10(wavc[i]->gain));
  printf(" %9f",wavc[i]->score);
  printf(" %9f\r\n",wavc[i]->max_amp);
 }
 free(wavc[0]->initials);
 wavc[0]->initials = NULL;
 //shortest_sample_num -= (biggest_delay-smallest_delay);
 usable_length += smallest_delay;
 printf("biggest_delay    : %d\r\n",biggest_delay);
 printf("smallest_delay   : %d\r\n",smallest_delay);
 printf("original max amp : %f\r\n",wavc[0]->max_amp);
 printf("global_gain      : %f\r\n",global_gain);
 printf("usable_length    : %d\r\n",usable_length);
 printf("original rate    : %d\r\n",wavc[0]->sample_rate);
 /*------------------------------------------------------------------*/
 /*     Output gain&offset adjusted sounds as raw                    */
 /*------------------------------------------------------------------*/
 if(!testmode){
  for(i=0;i<wav_num;i++){
   int bytes_per_sample = 8;
   if(wavc[i]->is_16bit)bytes_per_sample = 4;
   int the_byte_offset = wavc[i]->file_offset;
   the_byte_offset += (wavc[i]->offset-smallest_delay)*bytes_per_sample;
   fseek(wavc[i]->wav,the_byte_offset,SEEK_SET);
   FILE *rawout;
   char *sp = strstr(wav_path[i],".wav");
   if(sp!=NULL)sp[0] = '\0';
   if(output_format==OUTPUT_FORMAT_WAV){
    strcat(wav_path[i],".final.wav");
   }else{
    strcat(wav_path[i],".raw");
   }
   rawout = fopen(wav_path[i],"wb");
   if(rawout==NULL){
    printf("Output path %s is not available.\r\n",wav_path[i]);
    return 14;
   }
   double thegain = wavc[i]->gain*32768.0*global_gain;
   int clipped = 0;
   if(output_format==OUTPUT_FORMAT_WAV){
    fputs("RIFF",rawout);
    int intbuf = 44-8+usable_length*4;
    fwrite(&intbuf,4,1,rawout);
    fputs("WAVE",rawout);
    fputs("fmt ",rawout);
    intbuf = 16;
    fwrite(&intbuf,4,1,rawout);
    intbuf = 65536*2+1;
    fwrite(&intbuf,4,1,rawout);
    intbuf = wavc[i]->sample_rate;
    fwrite(&intbuf,4,1,rawout);
    intbuf *= 2*2;
    fwrite(&intbuf,4,1,rawout);
    intbuf = 65536*16+2*2;
    fwrite(&intbuf,4,1,rawout);
    fputs("data",rawout);
    intbuf = usable_length*4;
    fwrite(&intbuf,4,1,rawout);
   }
   if(wavc[i]->is_float){
     for(j=0;j<usable_length*2;j++){
       float flobuf;
       short shobuf;
       fread(&flobuf,4,1,wavc[i]->wav);
       int ibtg = ((int)(((double)flobuf*thegain)+40000.5))-40000;
       if(ibtg<-32440 || 32440<ibtg){//2^15 * 0.99
        clipped++;
        if(ibtg<-32767)ibtg = -32767;
        if(ibtg>32767)ibtg = 32767;
       }
       int write_ok;//Big endian to little endian
       shobuf = ibtg;//(ibtg>>8) | (ibtg<<8);
       write_ok = fwrite(&shobuf,2,1,rawout);
       if(write_ok==0){
        printf("%s Write error: Disk is full.\r\n",wavc[i]->thumbnail);
        return 15;
       }
     }
   }else{//Integer wav format
     for(j=0;j<usable_length*2;j++){
       double dbuf;
       int intbuf;
       short shobuf;
       if(wavc[i]->is_16bit){
         fread(&shobuf,2,1,wavc[i]->wav);
         dbuf = shobuf/32768.0;
       }
       else{
         fread(&intbuf,4,1,wavc[i]->wav);
         dbuf = intbuf/2147483648.0;//32bit 2^31
       }
       int ibtg = ((int)((dbuf*thegain)+40000.5))-40000;
       if(ibtg<-32440 || 32440<ibtg){//2^15 * 0.99
         clipped++;
         if(ibtg<-32767)ibtg = -32767;
         if(ibtg>32767)ibtg = 32767;
       }
    
       /* char cbuf;
       int write_ok;
       cbuf = ibtg&0xFF;
       write_ok = fwrite(&cbuf,1,1,rawout);
       if(write_ok==0){
        printf("%s Write error: Disk is full.\r\n",wavc[i]->thumbnail);
        return 15;
       }
       cbuf = ibtg>>8;
       write_ok = fwrite(&cbuf,1,1,rawout);
       if(write_ok==0){
        printf("%s Write error: Disk is full.\r\n",wavc[i]->thumbnail);
        return 16;
       }*/
     
       int write_ok;//Big endian to little endian
       shobuf = ibtg;//(ibtg>>8) | (ibtg<<8);
       write_ok = fwrite(&shobuf,2,1,rawout);
       if(write_ok==0){
        printf("%s Write error: Disk is full.\r\n",wavc[i]->thumbnail);
        return 15;
       }
     }
   }

   if(clipped){
    printf("%s Clipped %d times.\r\n",wavc[i]->thumbnail,clipped);
   }
   fclose(wavc[i]->wav);
   fclose(rawout);
  }// end of for
 }// end of !testmode
 return 0;
}




 /*
 FILE *output_wav;
 output_wav = fopen("test_originial.wav","wb");
 if(output_wav==NULL){
  printf("Output path is not available.\r\n");
  return 10;
 }
 fputs("RIFF",output_wav);
 intbuf = 44-8+INITIAL_MAX*8;
 fwrite(&intbuf,4,1,output_wav);
 fputs("WAVE",output_wav);
 fputs("fmt ",output_wav);
 intbuf = 16;
 fwrite(&intbuf,4,1,output_wav);
 intbuf = 65536*2+1;
 fwrite(&intbuf,4,1,output_wav);
 intbuf = input_sampling_rate;
 fwrite(&intbuf,4,1,output_wav);
 intbuf *= 2*4;
 fwrite(&intbuf,4,1,output_wav);
 intbuf = 65536*32+2*4;
 fwrite(&intbuf,4,1,output_wav);
 fputs("data",output_wav);
 intbuf = INITIAL_MAX*8;
 fwrite(&intbuf,4,1,output_wav);

 for(i=0;i<INITIAL_MAX;i++){
  for(j=0;j<2;j++){
   int s = wavc[0]->initials[i];
   fwrite(&s,4,1,output_wav);
  }
 }
 fclose(output_wav);
 */


